(* ::Package:: *)

(* :Context: QuaternionAnalysis` *)

(* :Title: QuaternionAnalysis *)

(* :Author: Maria Irene Falcao, Fernando Miranda *)

(* :Version: Mathematica 10.0 *)

(* :Package Version: 1.0.0 *)

(* :Keywords:
    Fueter, Hamilton, quaternions, paravectors, monogenic functions
*)
(* ::Webpage: w3.math.uminho.pt/QuaternionsAnalysis *)

(* :History:
Version 0.1.0  June 2011
Version 1.0.0  November 2014
*)

(* :Copyright:  *)

(* :Requirements: Quaternions Package*)

(* :Warnings:

    Adds functionality to the following functions:
    Plus, Times, Divide, Power, Re, Conjugate, Dot, Abs, Norm, Sign, Derivative.

    In this package, the quaternions can have real valued or symbolic entries.

	Some of the new functions overshadow those of the Quaternions package:
	Abs, Norm, ScalarQ, Quaternions`Private`extfunc.
	
*)

(* :Summary:

	QuaternionAnalysis package adds functionalities to the Quaternions package 
	which implements Hamilton's quaternion algebra. It also provides tools for 
	manipulating regular quaternion valued functions, in the sense of Fueter.
	Some of the added new features include the possibility of performing 
	operations on functions defined in R^(n+1), n>=2. 

*)

(* :Sources:

	Falcao, M.I., Miranda, F.: Quaternions: A Mathematica package for 
	quaternionic analysis. Lecture Notes in Computer Science, 6784, 
	200-214 (2011)

	Fueter, R.: Die Funktionentheorie der Differetialgleichungen \Delta u = 0 
	und \Delta\Delta u = 0 mit vier reellen Variablen. Comm. Math. Helv. (7), 
	307-330 (1934-35)
	
	Guerlebeck, K., Habetha, K., Sproessig, W.: Holomorphic functions in the 
	plane and n-dimensional space. Birkhauser Verlag, Basel (2008)
	
	Sudbery, A.: Quaternionic analysis. Math. Proc. Camb. Phil. Soc. 85, 
	199-225 (1979)

*)

BeginPackage["QuaternionAnalysis`",{"Quaternions`"}]

If[Not@ValueQ[AbsVec::usage],AbsVec::usage=
"AbsVec[x] gives the absolute value of the vector part of x."]

If[Not@ValueQ[CauchyRiemannL::usage],CauchyRiemannL::usage =
"CauchyRiemannL[f] applies the Cauchy-Riemann operator to f, form the left."]

If[Not@ValueQ[CauchyRiemannR::usage],CauchyRiemannR::usage =
"CauchyRiemannR[f] applies the Cauchy-Riemann operator to f, form the right."]

If[Not@ValueQ[Ck::usage],Ck::usage =
"Ck[k,n] gives the alternating sum of Tk[k,s,n], i.e. \
Ck[k,n]= Sum[(-1)^sTks[k,s,n],{s,0,k}].
Ck[k]=Ck[k,$Dim-1]."]

If[Not@ValueQ[ComplexLike::usage],ComplexLike::usage =
"ComplexLike[x,y] represents a quaternion or a paravector by its real part \
x and the norm of its vector part y."];

If[Not@ValueQ[ComplexLike::argx],ComplexLike::argx =
"ComplexLike called with `1` argument(s); 2 arguments are expected."];

If[Not@ValueQ[ComplexToQuaternion::usage],ComplexToQuaternion::usage=
"ComplexToQuaternion[x,y] gives the quaternion x+yJ."]

If[Not@ValueQ[DiracL::usage],DiracL::usage =
"DiracL[f] applies the Dirac operator to f, from the left."]

If[Not@ValueQ[DiracR::usage],DiracR::usage =
"DiracR[f] applies the Dirac operator to f, form the right."]

If[Not@ValueQ[Laplace::usage],Laplace::usage =
"Laplace[f] applies the Laplace operator to f."]

If[Not@ValueQ[LeftMonogenicQ::usage],LeftMonogenicQ::usage =
"LeftMonogenicQ[f] gives True if f is left monogenic, i.e. CauchyRiemannL[f]=0."]

If[Not@ValueQ[MonogenicQ::usage],MonogenicQ::usage =
"MonogenicQ[f] gives True if f is monogenic, i.e. \
CauchyRiemannL[f]=CauchyRiemannR[f]=0."]

If[Not@ValueQ[Paravector::usage],Paravector::usage =
"Paravector[x0,x1,...xn] represents the paravector x0 + x1 e1 + ... + xn en, \
where {e1,...,en} is an orthonormal basis of the euclidean space R^n, \
with a noncomutative product according to the rules ei ej = -ej ei, ei^2=-1."];

If[Not@ValueQ[Pk::usage], Pk::usage =
"Pk[k,n,x] gives the monogenic polynomial of degree k, generalizing \
the complex powers.
If x is a quaternion (n=3) or a 3D-paravector (n=2), \
Pk[k,x]=Sum[Tks[k,s,n]x^(k-s)Conjugate[x]^s,{s,0,k}].
If x=ComplexLike[a,b]=a+\[Omega]b, \
Pk[k,n,x]=Sum[Binomial[k,s] a^(k-s) (\[Omega]b)^s Ck[s,n],{s,0,k}]."];

If[Not@ValueQ[PolarForm::usage],PolarForm::usage=
"PolarForm[expr] gives de polar form of expr."]

If[Not@ValueQ[PureQuaternionQ::usage],PureQuaternionQ::usage =
"PureQuaternionQ[x] gives True if the quaternion x is a pure quaternion, \
i.e. x = Quaternion[0,x1,x2,x3]"];

If[Not@ValueQ[PureQuaternionQ::argx],PureQuaternionQ::argx =
"PureQuaternion argument must be a quaternion."];

If[Not@ValueQ[QPower::usage],QPower::usage =
"QPower[expr,n]  a recursive definition of expr^n."]

If[Not@ValueQ[QuaternionToComplex::usage],QuaternionToComplex::usage=
"QuaternionToComplex[x] gives a complex representation of the quaternion x."]

If[Not@ValueQ[QuaternionToComplexMatrix::usage],QuaternionToComplexMatrix::usage=
"QuaternionToComplexMatrix[x] gives a complex representation matrix of the quaternion x."]

If[Not@ValueQ[QuaternionToMatrixL::usage],QuaternionToMatrixL::usage=
"QuaternionToMatrixL[x] gives a real left representation matrix of the quaternion x."]

If[Not@ValueQ[QuaternionToMatrixR::usage],QuaternionToMatrixR::usage=
"QuaternionToMatrixR[x] gives a real right representation matrix of the quaternion x."]

If[Not@ValueQ[R::usage],R::usage=
"R is the default variable representing the second component of a \
ComplexLike obbject."];

If[Not@ValueQ[RightMonogenicQ::usage],RightMonogenicQ::usage =
"RightMonogenicQ[f] gives True if f is right monogenic, i.e. CauchyRiemannR[f]=0"]

If[Not@ValueQ[SetCoordinates::usage],SetCoordinates::usage = 
"SetCoordinates[CoordinatesSequence] sets the default \
coordinates to be CoordinatesSequence.
SetCoordinates[a,b,c] sets the default coordinates to be a, b and c.
SetCoordinates[5] sets the default coordinates to be X0, X1, X2, X3 and X4."]

If[Not@ValueQ[SetCoordinates::valid],SetCoordinates::valid =
"The coordinates system is set to `1`."];

If[Not@ValueQ[SymmetricProduct::usage],SymmetricProduct::usage =
"SymmetricProduct[x,y] gives de symmetric product of the quaternions x and y."];

If[Not@ValueQ[Tks::usage],Tks::usage =
"Tks[k,s,n] gives the (k,s) element of a triangle associated with Pk.
Tks[k,s]=Tks[k,s,$Dim-1]."];

If[Not@ValueQ[ToComplexLike::usage],ToComplexLike::usage=
"ToComplexLike[x] represents x in the complex-like form."];

If[Not@ValueQ[Vec::usage],Vec::usage =
"Vec[x] gives the vector part of the object x."];

If[Not@ValueQ[W::usage],W::usage=
"W[x] gives the sign of the vector part of x, i.e. W[x]=Vec[x]/Abs[Vec[x]]."]

If[Not@ValueQ[$CoordinatesList::usage],$CoordinatesList::usage =
"$CoordinatesList is the list of variables. {X0,X1,X2,X3} is the default \
coordinates list."];

If[Not@ValueQ[$CoordinatesList::invalidDim],$CoordinatesList::invalidDim =
"The space dimension must be at least 2.\ 
The coordinates list has not been changed and still is `1`."];

If[Not@ValueQ[$CoordinatesList::invalid],$CoordinatesList::invalid =
"Invalid argument.\ 
The coordinates list has not been changed and still is `1`."];

If[Not@ValueQ[$Dim::usage],$Dim::usage =
"$Dim is the dimension of the space. By default the space dimension is 4. \ 
The $Dim value is automatically adjusted by the SetCoordinates function."]

Which[StringPosition[ScalarQ::usage, "Quaternions"] === {},
ScalarQ::usage = StringJoin[ScalarQ::usage, "\n This function overshadows  \
the function  defined in the Quaternions package in order to allow symbolic entries."]]

Begin["`Private`"]

(* ********** Redefinition of ScalarQ ********** *)

MyScalar = {
    Complex[x_, y_] :> False,
    ComplexLike[x_, y_] :> False,
    Plus[x_, Times[Complex[0, 1], y_]]:> False,
    Plus[Complex[0, x_], y_]:> False,
    Times[Complex[0, 1], y_]:> False,
    Times[Complex[0, x_], y_]:> False,
    x_ :> Head[N[x]] =!= Complex && Head[N[x]] =!= Quaternion && Head[N[x]] =!= Paravector};



ScalarQ[q_] := q /. MyScalar

(* fix bug on Quaternions package: see fix bug 66775 -- charlesp *)

Quaternions`Private`extfunc[func_, a_, b_] :=
    Re[b] + AbsIJK[b]*Quaternions`Private`SignIJK[a]
    

(* ==================== SetCoordinates ========================== *)

SetCoordinates[]:= Print[$CoordinatesList];

SetCoordinates[1] := Message[$CoordinatesList::invalidDim,$CoordinatesList];

SetCoordinates[n_?(IntegerQ@# && # > 1 &)] := Module[{cl}, 
     cl = Table[ToExpression[StringJoin[ToString[QuaternionAnalysis`X], ToString[i]]], {i, 0,n - 1}];
     If[(Head /@ cl // Union) =!= {Symbol}, Apply[SetCoordinates, cl],
        (Apply[Unprotect, $CoordinatesList]; $CoordinatesList = cl;
         Apply[Protect, $CoordinatesList];Message[SetCoordinates::valid, $CoordinatesList];
         Unprotect[$Dim]; $Dim=n;Protect[$Dim];
         $Assumptions=Union[$Assumptions&&Apply[And,Map[Element[#,Reals]&,$CoordinatesList]]];)
       ]];

SetCoordinates[_Symbol] := Message[$CoordinatesList::invalidDim,$CoordinatesList];

SetCoordinates[coord__Symbol] := (Apply[Unprotect, $CoordinatesList];
        $CoordinatesList = {coord}; Protect[coord];
        Message[SetCoordinates::valid, {coord}];
        Unprotect[$Dim]; $Dim=Length@{coord};Protect[$Dim];
        $Assumptions=Union[$Assumptions&&Apply[And,Map[Element[#,Reals]&,$CoordinatesList]]];)

SetCoordinates[__] := Message[$CoordinatesList::invalid, $CoordinatesList];

(* *************** PureQuaternionQ *************** *)

PureQuaternionQ[_]:=Message[PureQuaternion::argx]

PureQuaternionQ[q_Quaternion] := q[[1]] === 0

SyntaxInformation[PureQuaternionQ] = {"ArgumentsPattern" -> {_}};

(* *************** ComplexLike *************** *)

 ComplexLike[]:= Message[ComplexLike::argx, 0];

 ComplexLike[_]:= Message[ComplexLike::argx, 1];

 ComplexLike[_, _, n__]:= Message[ComplexLike::argx, 2 + Length[{n}]];
 
 SyntaxInformation[ComplexLike] = {"ArgumentsPattern" -> {_, _}};

(* *************** ToComplexLike *************** *)

ToComplexLike[a_Quaternion|a_Paravector]:= ComplexLike[First[a], AbsVec[a]];

SyntaxInformation[ToComplexLike]={"ArgumentsPattern" -> {_}};

(* *************** TraditionalForm *************** *)

Quaternion /:
 TraditionalForm[Quaternion[a_, b_, c_, d_]]:= a + b  System`\[DoubleStruckI] + c  System`\[DoubleStruckJ] + d  System`\[DoubleStruckK]

ComplexLike /:
 TraditionalForm[ComplexLike[a_, b_]] := a + System`\[DoubleStruckW] b

Paravector /:
 TraditionalForm[a_Paravector] := Dot[Apply[List, a], Table[Subscript[Global`e, i], {i, 0, Length[a] - 1}]]

(* *************** Rules for Plus *************** *)

Paravector /:
    Plus[a_Paravector, b_Paravector] := Apply[Paravector, Apply[List, a] + Apply[List, b]]

Paravector /:
    Plus[a_?ScalarQ, b_Paravector] := ReplacePart[b, 1 -> b[[1]] + a]

ComplexLike /:
    ComplexLike[a_, b_] + ComplexLike[c_, d_]:= ComplexLike[a + c, b + d]

(* *************** Rules for Times *************** *)

ComplexLike /:
     x_?ScalarQ * ComplexLike[a_, b_] := ComplexLike[x a, x b]

ComplexLike /:
 ComplexLike[a_, b_]*ComplexLike[c_, d_] := ComplexLike[a c - b d, a d + b c]

Paravector /: Times[a_?ScalarQ, b_Paravector] := Apply[Paravector, Times[a, Apply[List, b]]]

(* *************** Symmetric Product *************** *)

SymmetricProduct[a:Quaternion[__?ScalarQ],b:Quaternion[__?ScalarQ]]:=(a**b+b**a)/2

SyntaxInformation[SymmetricProduct]={"ArgumentsPattern" -> {_,_}};

(* *************** Rules for simple functions *************** *)

Paravector /:
    Conjugate[a_Paravector] :=  2 Re[a] - a

ComplexLike /:
	Conjugate[ComplexLike[a_, b_]]:=ComplexLike[a,-b]
	
Quaternion /:
	Norm[a:Quaternion[__?ScalarQ]] := Sqrt[Apply[Plus, (Apply[List, a])^2]]

ComplexLike /:
    Norm[a:ComplexLike[_?ScalarQ,_?ScalarQ]] := Sqrt[Apply[Plus, (Apply[List, a])^2]]

Paravector /:
    Norm[a:Paravector[__?ScalarQ]] := Sqrt[Apply[Plus, (Apply[List, a])^2]]

ComplexLike /:
     Abs[a:ComplexLike[_?ScalarQ,_?ScalarQ]] := Norm[a]

Paravector /:
     Abs[a:Paravector[__?ScalarQ]] :=  Norm[a]
     
(* Rewriting of the Quaternions package functions affected by redefinition of the norm *)

Quaternion /:
    Abs[  a:Quaternion[__?ScalarQ]]:= Norm[a]

Quaternion /:
    EvenQ[a:Quaternion[__?ScalarQ]]:= EvenQ[Norm[a]^2]

Quaternion /:
    OddQ[ a:Quaternion[__?ScalarQ]]:= OddQ[Norm[a]^2]
    
Quaternions`Private`quatprime[x_]:= If[IntegerQ[x],False,PrimeQ[Norm[x /. Quaternions`Private`toquat]]]

Quaternion /:
    Power[a:Quaternion[__?ScalarQ],-1]:= Conjugate[a] ** 1/Norm[a]^2
    
Quaternion /:
    Divide[a:Quaternion[__?ScalarQ], b:Quaternion[__?ScalarQ]]:=
    a ** (Conjugate[b] ** 1/Norm[b]^2)

(* *************** Re *************** *)

ComplexLike /:
    Re[a : ComplexLike[_?ScalarQ,_?ScalarQ]] := First[a]

Paravector /:
    Re[a : Paravector[__?ScalarQ]] := First[a]

(* *************** Vec *************** *)

Vec[a:Quaternion[__?ScalarQ]] := a-Re[a]

Vec[a:Quaternion[__?ScalarQ]] := a - Re[a]

Vec[a:Paravector[__?ScalarQ]] := a - Re[a]

Vec[a:ComplexLike[_?ScalarQ,_?ScalarQ]] := ComplexLike[0, a[[2]]]

SyntaxInformation[Vec] = {"ArgumentsPattern" -> {_}};

(* *************** AbsVec *************** *)

AbsVec[a_Quaternion | a_Paravector] :=  Abs[Vec[a]]

SyntaxInformation[AbsVec] = {"ArgumentsPattern" -> {_}};

(* *************** W *************** *)

W[Quaternion[_,0,0,0]]:=Quaternion[0,0,0,0]

W[a:Paravector[_,0...]]:=0 a

W[a_Quaternion|a_Paravector]:= Vec[a]/AbsVec[a]

SyntaxInformation[W] = {"ArgumentsPattern" -> {_}};

(* *************** Sign *************** *)

Quaternion /:
    Sign[Quaternion[0,0,0,0]]:= 0

(* *************** Dot *************** *)

Quaternion /:
	Dot[a_Quaternion,b_Quaternion]:=Apply[Dot,Apply[List,{a,b},{1}]]
	
Paravector /:
	Dot[a_Paravector,b_Paravector]:=Apply[Dot,Apply[List,{a,b},{1}]]

(* ***************  Representation functions  *************** *)

PolarForm[Quaternion[0,0,0,0]]:={0,Indeterminate}

PolarForm[Quaternion[a_?Positive,0,0,0]]:={a,0}

PolarForm[Quaternion[a_?Negative,0,0,0]]:={-a,Pi}

PolarForm[Paravector[0..]]:={0,Indeterminate}

PolarForm[Paravector[a_?Positive,0...]]:={a,0}

PolarForm[Paravector[a_?Negative,0...]]:={-a,Pi}

PolarForm[Paravector[a_?Positive,0..]]:={a,0}

PolarForm[a_Quaternion|a_Paravector]:=List[Abs[a],ArcCot[Re[a]/AbsVec[a]]]

SyntaxInformation[PolarForm] = {"ArgumentsPattern" -> {_}};

QuaternionToMatrixL[Quaternion[a_?ScalarQ,b_?ScalarQ,c_?ScalarQ,d_?ScalarQ]]:=
	{{a,-b,-c,-d},{b,a,-d,c},{c,d,a,-b},{d,-c,b,a}}
	
SyntaxInformation[QuaternionToMatrixL] = {"ArgumentsPattern" -> {_}};

QuaternionToMatrixR[Quaternion[a_?ScalarQ,b_?ScalarQ,c_?ScalarQ,d_?ScalarQ]]:=
	{{a,-b,-c,-d},{b,a,d,-c},{c,-d,a,b},{d,c,-b,a}}

SyntaxInformation[QuaternionToMatrixR] = {"ArgumentsPattern" -> {_}};

QuaternionToComplexMatrix[Quaternion[a_?ScalarQ,b_?ScalarQ,c_?ScalarQ,d_?ScalarQ]]:=
	{{a+b I,c+d I},{-c+d I,a-b I}}
	
SyntaxInformation[QuaternionToComplexMatrix] = {"ArgumentsPattern" -> {_}};

QuaternionToComplex[Quaternion[a_?ScalarQ,b_?ScalarQ,c_?ScalarQ,d_?ScalarQ]]:={a+b I, c+d I}

SyntaxInformation[QuaternionToComplex] = {"ArgumentsPattern" -> {_}};

ComplexToQuaternion[a_,b_]:=Quaternion[Re[a],Im[a],Re[b],Im[b]]

SyntaxInformation[ComplexToQuaternion] = {"ArgumentsPattern" -> {_,_}};

(* *************** QPower *************** *)
 (* A recursive definition of the Power *)

QPower[q_Quaternion, 0] = Quaternion[1,0,0,0];

QPower[q_Quaternion, 1] = q;

QPower[q_Quaternion, n_Integer] := q ** QPower[q, n - 1];

SyntaxInformation[QPower] = {"ArgumentsPattern" -> {_,_}};

(* =============== Differential Operators =============== *)

If[Not@ValueQ[DiffOp::WrongDim],DiffOp::WrongDim = "The dimension space is `1`; \
to apply the `2` operator to a Quaternion, the dimension space must be at least 4."]

QuatProductMatrix=
{{{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1}},
{{0, -1, 0, 0}, {1, 0, 0, 0}, {0, 0, 0, -1}, {0, 0, 1, 0}},
{{0, 0, -1, 0}, {0, 0, 0, 1}, {1, 0, 0, 0}, {0, -1, 0, 0}},
{{0, 0, 0, -1}, {0, 0, -1, 0}, {0, 1, 0, 0}, {1, 0, 0, 0}}};

CauchyRiemannL[Quaternion[f0_, f1_, f2_, f3_]] :=
 Which[
       $Dim < 4,Message[DiffOp::WrongDim,$Dim,CauchyRiemannR],
       True,MapAll[Simplify,Apply[Quaternion,Inner[D[#2,#1]&,$CoordinatesList,QuatProductMatrix.{f0,f1,f2,f3}]]]
      ]

SyntaxInformation[CauchyRiemannL] = {"ArgumentsPattern" -> {_}};

CauchyRiemannR[Quaternion[f0_, f1_, f2_, f3_]] :=
 Which[
       $Dim < 4,Message[DiffOp::WrongDim,$Dim,CauchyRiemannL],
       True,MapAll[Simplify,Apply[Quaternion,Inner[D[#1,#2]&,{f0,f1,f2,f3}.QuatProductMatrix,$CoordinatesList]]]
      ]

SyntaxInformation[CauchyRiemannR] = {"ArgumentsPattern" -> {_}};

DiracL[Quaternion[f0_, f1_, f2_, f3_]] :=
 Which[
       $Dim < 4,Message[DiffOp::WrongDim,$Dim,CauchyRiemannR],
       True,MapAll[Simplify,Apply[Quaternion,Inner[D[#2,#1]&,ReplacePart[$CoordinatesList,1->Null],QuatProductMatrix.{f0,f1,f2,f3}]]]
      ]

SyntaxInformation[DiracL] = {"ArgumentsPattern" -> {_}};

DiracR[Quaternion[f0_, f1_, f2_, f3_]] :=
 Which[
       $Dim < 4,Message[DiffOp::WrongDim,$Dim,CauchyRiemannL],
       True,MapAll[Simplify,Apply[Quaternion,Inner[D[#1,#2]&,{f0,f1,f2,f3}.QuatProductMatrix,ReplacePart[$CoordinatesList,1->Null]]]]
      ]

SyntaxInformation[DiracR] = {"ArgumentsPattern" -> {_}};

Laplace[Quaternion[f0_, f1_, f2_, f3_]] :=
 Which[
       $Dim < 4,Message[DiffOp::WrongDim,$Dim,Laplace],
       True,Apply[Quaternion,Apply[Plus,Outer[D[#2, {#1, 2}] &,Take[$CoordinatesList, 4], {f0, f1, f2, f3}]]]
      ]

SyntaxInformation[Laplace] = {"ArgumentsPattern" -> {_}};

LeftMonogenicQ[q_Quaternion]:= CauchyRiemannL[q] === Quaternion[0, 0, 0, 0];

SyntaxInformation[LeftMonogenicQ] = {"ArgumentsPattern" -> {_}};

RightMonogenicQ[q_Quaternion]:= CauchyRiemannR[q] === Quaternion[0, 0, 0, 0];

SyntaxInformation[RightMonogenicQ] = {"ArgumentsPattern" -> {_}};

MonogenicQ[f_Paravector | f_Quaternion]:= Module[{Jac},
  Jac =Simplify[Outer[D[#2, #1] &,$CoordinatesList,Apply[List,f]]];
  Jac[[1]] = -Jac[[1]];
  Simplify@Tr[Jac] === 0 && Transpose[Jac] === Jac
  ]
  
Protect[R];

MonogenicQ[ComplexLike[u_, v_]]:=
 Simplify[D[u, First[$CoordinatesList]]-($Dim - 2) v/R - D[v, R]]===0 && 
  Simplify[D[v, First[$CoordinatesList]]+D[u, R]]===0

MonogenicQ[ComplexLike[u_, v_],X0_,R_]:=
 Simplify[D[u, X0]-($Dim - 2) v/R - D[v, R]]===0 && 
  Simplify[D[v, X0]+D[u, R]]===0

SyntaxInformation[MonogenicQ] = {"ArgumentsPattern" -> {_,_.,_.}};

Quaternion /:
 Derivative[f_Quaternion]:= If[MonogenicQ[f],
      Map[D[#,$CoordinatesList[[1]]]&,f],
      Message[MonogenicQ::Fail]]

Paravector /:
 Derivative[f_Paravector]:= If[MonogenicQ[f],
     Map[D[#,$CoordinatesList[[1]]] &,f],
     Message[MonogenicQ::Fail]]

ComplexLike /:
 Derivative[ComplexLike[u_,v_]]:= 
     If[MonogenicQ[ComplexLike[u,v]],
       ComplexLike[D[u,First[$CoordinatesList]],D[v,First[$CoordinatesList]]],
       Message[MonogenicQ::Fail]]

ComplexLike /:
 Derivative[ComplexLike[u_,v_],X0_,R_]:= 
     If[MonogenicQ[ComplexLike[u,v],X0,R],
       ComplexLike[D[u,X0],D[v,X0]],
       Message[MonogenicQ::Fail]]


(* *************** Basic Polynomials *************** *)

Tks[0, 0, _] := 1;

Tks[k_, s_, n_:($Dim-1)] := 
    Assuming[s \[Element] Integers && s >= 0 && k \[Element] Integers && s <= k,
   FullSimplify[k!/Pochhammer[n, k] (Pochhammer[(n + 1)/2, k - s] Pochhammer[(n - 1)/2,s])/((k - s)! s!)]];

SyntaxInformation[Tks] = {"ArgumentsPattern" -> {_,_,_.}};

Ck[k_, n_:($Dim-1)] := Sum[(-1)^s Tks[k, s, n],{s,0,k}]//Simplify;

SyntaxInformation[Ck] = {"ArgumentsPattern" -> {_,_.}};

NaturalQ = NonNegative[#] && IntegerQ[#] &;

Pk[k_?NaturalQ, x_Quaternion] := Sum[Tks[k, s] QPower[x, k - s]**QPower[Conjugate[x], s],{s,0,k}] // Simplify

Pk[k_?NaturalQ,x_Paravector] := Module[{y = Quaternion[x[[1]], x[[2]], x[[3]], 0]},

     Apply[Paravector, Delete[Simplify[Sum[Tks[k, s, 2] QPower[y, k - s]** QPower[Conjugate[y], s],{s,0,k}]], 4]]]

Pk[k_?NaturalQ, n_?NaturalQ, x_ComplexLike] :=  
     Apply[ComplexLike, Simplify[Through[{Re, Im}[Sum[Binomial[k, s] x[[1]]^(k - s) Ck[s, n] (I x[[2]])^s,{s,0,k}]]]]]

Pk[k_?NaturalQ, n_?NaturalQ, ComplexLike[0, b_]] :=Apply[ComplexLike, Simplify[Through[{Re, Im}[Ck[k, n] (I b)^k]]]]

Pk[k_?NaturalQ, n_?NaturalQ, ComplexLike[a_, 0]] :=ComplexLike[a^k, 0]

SyntaxInformation[Pk] = {"ArgumentsPattern" -> {_,_,_.}};

(* *************** Error messages *************** *)

MonogenicQ::Fail = "The function is non monogenic."

End[]  (* QuaternionAnalysis`Private`*)

(* *************** Defaults *************** *)

SetCoordinates[4];

$Assumptions = $Assumptions&&R>=0;

(* ******** updating verification ********* *)

Quiet[$LastVersion=URLFetch["http://w3.math.uminho.pt/QuaternionAnalysis/version"]];
Which[$LastVersion===$Failed,Null,
	(# != "1.0.0" && StringPosition[#, "not found"] == {}) & @ $LastVersion
	,
	MessageDialog[URLFetch["http://w3.math.uminho.pt/QuaternionAnalysis/out-of-date"]]]
	
Remove[$LastVersion];

EndPackage[] 

(* QuaternionAnalysis`*)